import copy
import logging
from unittest.mock import patch, MagicMock

import pytest

from perun.connector import perun_openapi
from perun.connector.adapters.PerunRpcAdapter import PerunRpcAdapter
from perun.connector.models.MemberStatusEnum import MemberStatusEnum
from perun.connector.models.Facility import Facility
from perun.connector.models.Group import Group
from perun.connector.models.Member import Member
from perun.connector.models.Resource import Resource
from perun.connector.models.User import User
from perun.connector.models.UserExtSource import UserExtSource
from perun.connector.models.VO import VO
from perun.connector.perun_openapi import ApiException
from perun.connector.utils.ConfigStore import ConfigStore


class HttpResponse:
    def __init__(self, input_data: str):
        self.status = None
        self.data = input_data
        self.reason = None

    def getheaders(self):
        return "headers"


ADAPTER = PerunRpcAdapter(ConfigStore.get_openapi_config(), ConfigStore.get_attribute_map())

# sample groups - external representation (Devel)
TEST_GROUP_EXTERNAL_REPRESENTATION_1 = {
    "id": 1,
    "vo_id": 10,
    "uuid": "sample-uuid-value",
    "name": "sample:group:name",
    "description": "This is a sample group",
}

TEST_GROUP_EXTERNAL_REPRESENTATION_2 = {
    "id": 2,
    "vo_id": 10,
    "uuid": "sample-uuid-value-2",
    "name": "sample:group:name:specific",
    "description": "This is a sample sub-group",
}

# sample groups - internal representation

TEST_VO = VO(62, "CESNET e-infrastruktura", "einfra")
TEST_MEMBER_INTERNAL_REPRESENTATION = Member(5, TEST_VO, "VALID")
TEST_MEMBER_EXTERNAL_REPRESENTATION = {"id": 5, "vo_id": 1, "status": "VALID"}
TEST_USER = User(10, "John Doe")
SAMPLE_SHORT_GROUP_NAME = "short_name"

TEST_GROUP_INTERNAL_REPRESENTATION_1 = Group(
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["id"],
    TEST_VO,
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["uuid"],
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["name"],
    f"{SAMPLE_SHORT_GROUP_NAME}:"
    f'{TEST_GROUP_EXTERNAL_REPRESENTATION_1["name"]}',
    TEST_GROUP_EXTERNAL_REPRESENTATION_1["description"],
)

TEST_GROUP_INTERNAL_REPRESENTATION_2 = Group(
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["id"],
    TEST_VO,
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["uuid"],
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["name"],
    f"{SAMPLE_SHORT_GROUP_NAME}:"
    f'{TEST_GROUP_EXTERNAL_REPRESENTATION_2["name"]}',
    TEST_GROUP_EXTERNAL_REPRESENTATION_2["description"],
)

# sample Facility
TEST_RP_IDENTIFIER_1 = "test rp identifier"
TEST_RP_IDENTIFIER_2 = "test rp identifier 2"

TEST_EXTERNAL_FACILITY_1 = {
    "id": 1,
    "name": "test name",
    "description": "test description",
}
TEST_EXTERNAL_FACILITY_2 = {
    "id": 2,
    "name": "test name 2",
    "description": "test description 2",
}

TEST_SINGLE_PERUN_FACILITY = [TEST_EXTERNAL_FACILITY_1]
TEST_MULTIPLE_PERUN_FACILITIES = [
    TEST_EXTERNAL_FACILITY_1,
    TEST_EXTERNAL_FACILITY_2,
]

TEST_INTERNAL_FACILITY_1 = Facility(
    TEST_EXTERNAL_FACILITY_1["id"],
    TEST_EXTERNAL_FACILITY_1["name"],
    TEST_EXTERNAL_FACILITY_1["description"],
    TEST_RP_IDENTIFIER_1,
)

TEST_INTERNAL_FACILITY_2 = Facility(
    TEST_EXTERNAL_FACILITY_2["id"],
    TEST_EXTERNAL_FACILITY_2["name"],
    TEST_EXTERNAL_FACILITY_2["description"],
    TEST_RP_IDENTIFIER_2,
)

MULTIPLE_ATTRS_ERROR_TEXT = (
    f"There is more than one facility with rpID '" f"{TEST_RP_IDENTIFIER_1}'."
)
NO_ATTRS_ERROR_TEXT = f"No facility with rpID '{TEST_RP_IDENTIFIER_1}' found."


def get_entity_specific_attributes(attribute_type: str):
    generated_attrs = []
    attr_names = []
    attr_values = []
    attr_namespaces = []

    if attribute_type == "user":
        attr_names = [
            "perunUserAttribute_phone",
            "perunUserAttribute_groupNames",
            "perunUserAttribute_preferredLanguage",
        ]
        attr_values = [
            None,
            ["group 1 name", "group 2 name"],
            "sample attribute value 3",
        ]
        attr_namespaces = [
            "urn:perun:user:attribute-def:def",
            "urn:perun:user:attribute-def:virt",
            "urn:perun:user:attribute-def:def",
        ]

    elif attribute_type == "vo":
        attr_names = ["perunVoAttribute_aup"]
        attr_values = ["sample value"]
        attr_namespaces = ["urn:perun:vo:attribute-def:def"]

    elif attribute_type == "entityless":
        attr_names = [
            "perunEntitylessAttribute_orgAups",
            "perunEntitylessAttribute_orgAups",
        ]
        attr_values = [None, "sample attribute value"]
        attr_namespaces = [
            "urn:perun:entityless:attribute-def:def",
            "urn:perun:entityless:attribute-def:def",
        ]

    elif attribute_type == "facility":
        attr_names = ["perunFacilityAttr_checkGroupMembership"]
        attr_values = ["test value"]
        attr_namespaces = ["urn:perun:facility:attribute-def:def"]

    for index, attr_name in enumerate(attr_names):
        generated_attrs.append(
            {
                "id": index,
                "name": attr_name,
                "namespace": attr_namespaces[index],
                "display_name": f"display name for {attr_name}",
                "type": "some type",
                "value": attr_values[index],
                "friendly_name": attr_name.split("_")[-1],
            }
        )

    return generated_attrs


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
def test_get_perun_user_with_middle_name_and_title_before_after(
    mock_request_1,
):
    user_with_middle_name_title_before_after = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": "Frederick",
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }
    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(  # noqa E501
        return_value=user_with_middle_name_title_before_after
    )

    expected_user = User(10, "Ing. John Frederick Doe Ph.D.")
    result_user = ADAPTER.get_perun_user("10", ["John Doe"])

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
def test_get_perun_user_without_middle_name_and_title_before_after(
    mock_request_1,
):
    user_without_middle_name_title_before_after = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }
    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(  # noqa E501
        return_value=user_without_middle_name_title_before_after
    )

    expected_user = User(10, "Ing. John Doe Ph.D.")
    result_user = ADAPTER.get_perun_user("10", ["John Doe"])

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
def test_user_without_middle_name_and_title_before(mock_request_1):
    user_without_middle_name_title_before = {
        "id": 10,
        "title_before": "Ing.",
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": None,
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(  # noqa E501
        return_value=user_without_middle_name_title_before
    )

    expected_user = User(10, "Ing. John Doe")
    result_user = ADAPTER.get_perun_user("10", ["John Doe"])

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
def test_user_without_middle_name_and_title_after(mock_request_1):
    user_without_middle_name_title_after = {
        "id": 10,
        "title_before": None,
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": "Ph.D.",
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(  # noqa E501
        return_value=user_without_middle_name_title_after
    )

    expected_user = User(10, "John Doe Ph.D.")
    result_user = ADAPTER.get_perun_user("10", ["John Doe"])

    assert result_user == expected_user


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
def test_user_without_middle_name_and_titles(mock_request_1):
    user_without_middle_name_and_titles = {
        "id": 10,
        "title_before": None,
        "first_name": "John",
        "middle_name": None,
        "last_name": "Doe",
        "title_after": None,
    }

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_by_ext_source_name_and_ext_login = MagicMock(  # noqa E501
        return_value=user_without_middle_name_and_titles
    )

    expected_user = User(10, "John Doe")
    result_user = ADAPTER.get_perun_user("10", ["John Doe"])

    assert result_user == expected_user


# get_get_member_groups tests
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi"
    ".get_all_member_groups"
)
@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
def test_get_member_groups_found_member_groups(
    mock_request_1, mock_request_2, mock_request_3, mock_request_4
):
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value={"value": SAMPLE_SHORT_GROUP_NAME}
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=TEST_VO
    )

    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_all_member_groups = MagicMock(  # noqa E501
        return_value=[
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            TEST_GROUP_EXTERNAL_REPRESENTATION_2,
        ]
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=TEST_MEMBER_INTERNAL_REPRESENTATION
    )

    result_groups = ADAPTER.get_member_groups(user=TEST_USER, vo=TEST_VO)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_groups_member_not_found(mock_request_1):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=None
    )

    result_groups = ADAPTER.get_member_groups(user=TEST_USER, vo=TEST_VO)
    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
@patch(
    "perun.connector.perun_openapi.api.resources_manager_api.ResourcesManagerApi"
    ".get_assigned_groups"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
def test_get_sp_groups_found_sp_groups(
    mock_request_1,
    mock_request_2,
    mock_request_3,
    mock_request_4,
):
    test_resources = [Resource(1, None, None, None)]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value={"value": SAMPLE_SHORT_GROUP_NAME}
    )

    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(  # noqa E501
        return_value=test_resources
    )

    perun_openapi.api.resources_manager_api.ResourcesManagerApi.get_assigned_groups = MagicMock(  # noqa E501
        return_value=[
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            TEST_GROUP_EXTERNAL_REPRESENTATION_2,
            TEST_GROUP_EXTERNAL_REPRESENTATION_2,
        ]
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=TEST_VO
    )

    result_groups = ADAPTER.get_sp_groups_by_facility(TEST_INTERNAL_FACILITY_1)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


def test_get_sp_groups_no_input_facility():
    result_groups = ADAPTER.get_sp_groups_by_facility(None)
    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
def test_get_sp_groups_no_resources_found(mock_request_1):
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(  # noqa E501
        return_value=[]
    )

    result_groups = ADAPTER.get_sp_groups_by_facility(TEST_INTERNAL_FACILITY_1)
    assert result_groups == []


@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch(
    "perun.connector.perun_openapi.api.groups_manager_api.GroupsManagerApi.get_group_by_name"
)
def test_get_group_by_name(mock_request_1, mock_request_2, mock_request_3):
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=TEST_VO
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value={"value": SAMPLE_SHORT_GROUP_NAME}
    )

    perun_openapi.api.groups_manager_api.GroupsManagerApi.get_group_by_name = (
        MagicMock(return_value=TEST_GROUP_EXTERNAL_REPRESENTATION_1)
    )

    result = ADAPTER.get_group_by_name(TEST_VO, "sample name")
    assert result == TEST_GROUP_INTERNAL_REPRESENTATION_1


@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name")
def test_get_vo_correct_arguments(mock_request_1, mock_request_2):
    test_vo_id = 1
    test_vo_short_name = "sample short name"
    test_vo = VO(test_vo_id, "sample name", test_vo_short_name)

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=test_vo
    )
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = (
        MagicMock(return_value=test_vo)
    )

    obtained_vo_by_id = ADAPTER.get_vo(vo_id=test_vo_id)
    obtained_vo_by_short_name = ADAPTER.get_vo(short_name=test_vo_short_name)

    assert test_vo == obtained_vo_by_id
    assert test_vo == obtained_vo_by_short_name


def test_get_vo_no_arguments():
    no_args_error_message = (
        "Neither short_name nor id was provided, "
        "please specify exactly one to find VO by."
    )

    with pytest.raises(ValueError) as error:
        _ = ADAPTER.get_vo()
    assert str(error.value.args[0]) == no_args_error_message


def test_get_vo_too_many_arguments():
    too_many_args_error_message = (
        "VO can be obtained either by its "
        "short_name or id, not both at the same "
        "time."
    )

    with pytest.raises(ValueError) as error:
        _ = ADAPTER.get_vo(vo_id=1, short_name="sample short name")
    assert str(error.value.args[0]) == too_many_args_error_message


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
def test_get_facility_by_rp_identifier_found_facility(mock_request_1):
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(  # noqa E501
        return_value=TEST_SINGLE_PERUN_FACILITY
    )

    result_facility = ADAPTER.get_facility_by_rp_identifier(
        TEST_RP_IDENTIFIER_1
    )

    assert result_facility == TEST_INTERNAL_FACILITY_1


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
def test_get_facility_by_rp_identifier_no_perun_attr_found(
    mock_request_1, caplog
):
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(  # noqa E501
        return_value=None
    )

    with caplog.at_level(logging.WARNING):
        result_facility = ADAPTER.get_facility_by_rp_identifier(
            TEST_RP_IDENTIFIER_1
        )
        assert result_facility is None
        assert NO_ATTRS_ERROR_TEXT in caplog.text


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_facilities_by_attribute"
)
def test_get_facility_by_rp_identifier_multiple_perun_attrs_found(
    mock_request_1, caplog
):
    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_facilities_by_attribute = MagicMock(  # noqa E501
        return_value=TEST_MULTIPLE_PERUN_FACILITIES
    )

    with caplog.at_level(logging.WARNING):
        result_facility = ADAPTER.get_facility_by_rp_identifier(
            TEST_RP_IDENTIFIER_1
        )
        assert result_facility is None
        assert MULTIPLE_ATTRS_ERROR_TEXT in caplog.text


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_groups_for_facility_where_user_is_active"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
def test_get_users_groups_on_facility_multiple_groups_found(
    mock_request_1, mock_request_2, mock_request_3
):
    perun_openapi.api.users_manager_api.UsersManagerApi.get_groups_for_facility_where_user_is_active = MagicMock(  # noqa E501
        return_value=[
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            TEST_GROUP_EXTERNAL_REPRESENTATION_2,
            TEST_GROUP_EXTERNAL_REPRESENTATION_2,
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
            TEST_GROUP_EXTERNAL_REPRESENTATION_1,
        ]
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value={"value": SAMPLE_SHORT_GROUP_NAME}
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=TEST_VO
    )

    result_groups = ADAPTER.get_users_groups_on_facility(
        TEST_INTERNAL_FACILITY_1, TEST_USER
    )

    for group in result_groups:
        assert isinstance(group, Group)

    assert [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
    ] == result_groups


def test_get_users_groups_on_facility_no_input_facility():
    result_groups = ADAPTER.get_users_groups_on_facility(None, TEST_USER)

    assert result_groups == []


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_groups_for_facility_where_user_is_active"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
@patch("perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id")
def test_get_users_groups_on_facility_no_groups_found(
    mock_request_1, mock_request_2, mock_request_3
):
    perun_openapi.api.users_manager_api.UsersManagerApi.get_groups_for_facility_where_user_is_active = MagicMock(  # noqa E501
        return_value=[]
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value={"value": SAMPLE_SHORT_GROUP_NAME}
    )

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_id = MagicMock(  # noqa E501
        return_value=TEST_VO
    )

    result_groups = ADAPTER.get_users_groups_on_facility(
        TEST_INTERNAL_FACILITY_1, TEST_USER
    )

    assert result_groups == []


# TODO add tests for more attributes when correct OpenAPI is generated
#  and make the tests work with searcher
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_facility_attributes_by_names"
)
@patch("perun.connector.perun_openapi.api.searcher_api.SearcherApi" ".get_facilities")
def test_get_facilities_by_attribute_value_correct_attribute(
    mock_request_1, mock_request_2
):
    perun_openapi.api.searcher_api.SearcherApi.get_facilities = MagicMock(  # noqa E501
        return_value=TEST_SINGLE_PERUN_FACILITY
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_facility_attributes_by_names = MagicMock(  # noqa E501
        return_value=[
            {
                "id": "test id",
                "name": "test name",
                "display_name": "test display name",
                "type": "test type",
                "value": TEST_INTERNAL_FACILITY_1.rp_id,
                "namespace": "urn:perun:facility:attribute-def:def",
                "friendly_name": "OIDCClientID",
            }
        ]
    )

    attribute = {"test name": "test value"}
    result_facilities = ADAPTER.get_facilities_by_attribute_value(attribute)

    assert result_facilities == [TEST_INTERNAL_FACILITY_1]


def test_get_facilities_by_attribute_value_empty_attribute(caplog):
    empty_attribute = {}
    wrong_number_of_attrs_error_text = (
        f"Attribute must contain exactly one "
        f"name and one value. Given attribute "
        f""
        f""
        f'contains: "{empty_attribute}".'
    )

    with caplog.at_level(logging.WARNING):
        result_facilities = ADAPTER.get_facilities_by_attribute_value(
            empty_attribute
        )
        assert result_facilities == []
        assert wrong_number_of_attrs_error_text in caplog.text


def test_get_facilities_by_attribute_value_too_many_attributes(caplog):
    multiple_attributes = {
        "test name": "test value",
        "test name 2": "test value 2",
    }
    wrong_number_of_attrs_error_text = (
        f"Attribute must contain exactly one "
        f"name and one value. Given attribute "
        f""
        f""
        f'contains: "{multiple_attributes}".'
    )

    with caplog.at_level(logging.WARNING):
        result_facilities = ADAPTER.get_facilities_by_attribute_value(
            multiple_attributes
        )
        assert result_facilities == []
        assert wrong_number_of_attrs_error_text in caplog.text


@patch(
    "perun.connector.perun_openapi.api.users_manager_api.UsersManagerApi"
    ".get_user_by_ext_source_name_and_ext_login"
)
@patch("perun.connector.adapters.PerunRpcAdapter.PerunRpcAdapter.get_perun_user")
def test_get_user_ext_source(mock_request_1, mock_request_2):
    source_id = 1
    source_login = "john@src"
    source_name = "sample source name"
    user_ext_source_external_representation = {
        "id": source_id,
        "login": source_login,
        "ext_source": {"name": source_name},
    }

    user_ext_source_internal_representation = UserExtSource(
        source_id, source_name, source_login, TEST_USER
    )

    perun_openapi.api.users_manager_api.UsersManagerApi.get_user_ext_source_by_ext_login_and_ext_source_name = MagicMock(  # noqa E501
        return_value=user_ext_source_external_representation
    )

    PerunRpcAdapter.get_perun_user = MagicMock(return_value=TEST_USER)

    result_ext_source = ADAPTER.get_user_ext_source(source_name, source_login)
    assert result_ext_source == user_ext_source_internal_representation


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_ext_source_attributes_by_names"
)
def test_get_user_ext_source_attributes_multiple_attributes(mock_request_1):
    test_attr_1_name = "perunUserAttribute_givenName"
    test_attribute_1 = {
        "id": 1,
        "name": test_attr_1_name,
        "namespace": "urn:perun:ues:attribute-def:def",
        "display_name": "sample attribute name 1",
        "type": "test type 1",
        "value": "sample attribute value 1",
        "friendly_name": "givenName",
    }

    test_attr_2_name = "perunUserAttribute_sn"
    test_attribute_2 = {
        "id": 2,
        "name": test_attr_2_name,
        "namespace": "urn:perun:ues:attribute-def:def",
        "display_name": "sample attribute name 2",
        "type": "test type 2",
        "value": "sample attribute value 2",
        "friendly_name": "sn",
    }

    attribute_names = [test_attr_1_name, test_attr_2_name]
    ext_source = UserExtSource(1, "sample name", "sample login", None)

    expected_attribute_1 = test_attribute_1.copy()
    expected_attribute_1.pop("namespace")

    expected_attribute_2 = test_attribute_2.copy()
    expected_attribute_2.pop("namespace")

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_ext_source_attributes_by_names = MagicMock(  # noqa E501
        return_value=[test_attribute_1, test_attribute_2]
    )

    result_attrs = ADAPTER.get_user_ext_source_attributes(
        ext_source, attribute_names
    )
    expected_attrs = {
        test_attr_1_name: expected_attribute_1,
        test_attr_2_name: expected_attribute_2,
    }

    assert result_attrs == expected_attrs


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_ext_source_attributes_by_names"
)
def test_get_user_ext_source_attributes_no_attributes(mock_request_1):
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_ext_source_attributes_by_names = MagicMock(  # noqa E501
        return_value=[]
    )

    ext_source = UserExtSource(1, "sample name", "sample login", None)
    result_attrs = ADAPTER.get_user_ext_source_attributes(ext_source, [])

    assert result_attrs == {}


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_status_by_user_and_vo_valid_member(mock_request_1):
    expected_status = "VALID"
    test_member_external_representation = {"id": 1, "status": expected_status}
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=test_member_external_representation
    )

    result_status = ADAPTER.get_member_status_by_user_and_vo(
        TEST_USER, TEST_VO
    )
    assert result_status == MemberStatusEnum(expected_status)


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_status_by_user_and_vo_invalid_member(mock_request_1):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"MemberNotExistsException"')
        )
    )

    result_status = ADAPTER.get_member_status_by_user_and_vo(
        TEST_USER, TEST_VO
    )
    assert result_status is None


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_status_by_user_and_vo_invalid_status(mock_request_1):
    invalid_status = "THIS_IS_NOT_A_VALID_STATUS"
    invalid_status_error_msg = f'"{invalid_status}" is not a valid state.'
    test_member_external_representation = {"id": 1, "status": invalid_status}

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=test_member_external_representation
    )

    with pytest.raises(ValueError) as error:
        _ = ADAPTER.get_member_status_by_user_and_vo(TEST_USER, TEST_VO)
        assert str(error.value.args[0]) == invalid_status_error_msg


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi" ".get_vo_by_short_name"
)
def test_is_user_in_vo_valid_member(mock_request_1, mock_request_2):
    valid_status_name = "VALID"
    test_valid_member_external_representation = {
        "id": 1,
        "status": valid_status_name,
    }

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = (
        MagicMock(E501return_value=TEST_VO)
    )

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=test_valid_member_external_representation
    )

    user_is_valid_member = ADAPTER.is_user_in_vo_by_short_name(TEST_USER, TEST_VO.short_name)
    assert user_is_valid_member


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi" ".get_vo_by_short_name"
)
def test_is_user_in_vo_invalid_member(mock_request_1, mock_request_2):
    invalid_status_name = "INVALID"
    test_valid_member_external_representation = {
        "id": 1,
        "status": invalid_status_name,
    }

    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = (
        MagicMock(return_value=TEST_VO)
    )

    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=test_valid_member_external_representation
    )

    user_is_valid_member = ADAPTER.is_user_in_vo_by_short_name(TEST_USER, TEST_VO.short_name)
    assert not user_is_valid_member


@patch(
    "perun.connector.perun_openapi.api.vos_manager_api.VosManagerApi" ".get_vo_by_short_name"
)
def test_is_user_in_vo_non_existing_vo(mock_request_1, caplog):
    perun_openapi.api.vos_manager_api.VosManagerApi.get_vo_by_short_name = (
        MagicMock(  # noqa E501
            side_effect=ApiException(
                http_resp=HttpResponse('"name":"VoNotExistsException"')
            )
        )
    )

    non_existing_short_name = "non_existing_short_name"
    vo_not_found_msg = (
        f'No VO with short name "{non_existing_short_name}" found'
    )

    with caplog.at_level(logging.DEBUG):
        user_is_valid_member = ADAPTER.is_user_in_vo_by_short_name(
            TEST_USER, non_existing_short_name
        )

        assert not user_is_valid_member
        assert vo_not_found_msg in caplog.text


def test_is_user_in_vo_user_without_id():
    user_without_id_msg = "User's ID is empty"
    user_without_id = User(None, "He, who shall not be named")

    with pytest.raises(ValueError) as error:
        _ = ADAPTER.is_user_in_vo_by_short_name(user_without_id, "some short name")
    assert str(error.value.args[0]) == user_without_id_msg


def test_is_user_in_vo_no_short_name_given():
    short_name_missing_msg = "VO short name is empty"
    missing_short_name = None

    with pytest.raises(ValueError) as error:
        _ = ADAPTER.is_user_in_vo_by_short_name(TEST_USER, missing_short_name)
    assert str(error.value.args[0]) == short_name_missing_msg


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_by_user_existing_user(mock_request_1):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        return_value=TEST_MEMBER_EXTERNAL_REPRESENTATION
    )

    result_member = ADAPTER.get_member_by_user(TEST_USER, TEST_VO)

    assert result_member == TEST_MEMBER_INTERNAL_REPRESENTATION


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_by_user_get_non_existing_user(mock_request_1, caplog):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"UserNotExistsException"')
        )
    )

    invalid_user_error_msg = (
        f' User with id "{TEST_USER.id}" does not exist in Perun.'
    )

    with caplog.at_level(logging.WARNING):
        result_member = ADAPTER.get_member_by_user(TEST_USER, TEST_VO)

        assert result_member is None
        assert invalid_user_error_msg in caplog.text


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_by_user_with_non_existing_vo(mock_request_1, caplog):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"VoNotExistsException"')
        )
    )

    invalid_vo_error_msg = (
        f'VO with id "{TEST_VO.id}" does not exist in Perun.'
    )

    with caplog.at_level(logging.WARNING):
        result_member = ADAPTER.get_member_by_user(TEST_USER, TEST_VO)

        assert result_member is None
        assert invalid_vo_error_msg in caplog.text


@patch(
    "perun.connector.perun_openapi.api.members_manager_api.MembersManagerApi"
    ".get_member_by_user"
)
def test_get_member_by_user_user_not_member_in_vo(mock_request_1, caplog):
    perun_openapi.api.members_manager_api.MembersManagerApi.get_member_by_user = MagicMock(  # noqa E501
        side_effect=ApiException(
            http_resp=HttpResponse('"name":"MemberNotExistsException"')
        )
    )

    invalid_vo_error_msg = (
        f'Member with VO "{TEST_VO.id}" and user id "{TEST_USER.id}" '
        f"does not "
        f"exist in Perun."
    )

    with caplog.at_level(logging.WARNING):
        result_member = ADAPTER.get_member_by_user(TEST_USER, TEST_VO)

        assert result_member is None
        assert invalid_vo_error_msg in caplog.text


@patch(
    "perun.connector.perun_openapi.api.facilities_manager_api.FacilitiesManagerApi"
    ".get_assigned_resources_for_facility"
)
@patch(
    "perun.connector.perun_openapi.api.resources_manager_api.ResourcesManagerApi"
    ".get_assigned_groups"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
def test_get_resource_capabilities(
    mock_request_1, mock_request_2, mock_request_3
):
    group_without_resource = Group(-2, None, "", "", "", "")
    test_user_groups = [
        TEST_GROUP_INTERNAL_REPRESENTATION_1,
        TEST_GROUP_INTERNAL_REPRESENTATION_2,
        group_without_resource,
    ]

    resource_without_group = {"id": -1}
    resource_of_group_1 = {"id": test_user_groups[0].id}
    resource_of_group_2 = {"id": test_user_groups[1].id}
    test_resource_groups = [
        [resource_of_group_1, resource_of_group_2],
        [resource_without_group],
        [],
    ]

    test_facility_resources = [{"id": 5}, {"id": 6}, {"id": 7}]

    perun_openapi.api.facilities_manager_api.FacilitiesManagerApi.get_assigned_resources_for_facility = MagicMock(  # noqa E501
        return_value=test_facility_resources
    )

    perun_openapi.api.resources_manager_api.ResourcesManagerApi.get_assigned_groups = MagicMock(  # noqa E501
        side_effect=test_resource_groups
    )

    resource_capabilities_of_groups = [
        "test capability 1",
        "test capability 2",
    ]
    resource_capabilities_without_group = ["test capability 3"]
    absent_capabilities = None
    external_capabilities = [
        {"value": resource_capabilities_of_groups},
        {"value": resource_capabilities_without_group},
        {"value": absent_capabilities},
    ]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        side_effect=external_capabilities
    )

    expected_capabilities = ["test capability 1", "test capability 2"]
    result_capabilities = ADAPTER.get_resource_capabilities_by_facility(
        TEST_INTERNAL_FACILITY_1, test_user_groups
    )

    assert sorted(result_capabilities) == sorted(expected_capabilities)


def test_get_resource_capabilities_no_input_facility():
    result_capabilities = ADAPTER.get_resource_capabilities_by_facility(None, [])
    assert result_capabilities == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
def test_get_facility_capabilities(mock_request_1):
    facility_capabilities = {
        "value": ["test capability 1", "test capability 2"]
    }
    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value=facility_capabilities
    )

    expected_capabilities = facility_capabilities["value"]
    result_capabilities = ADAPTER.get_facility_capabilities_by_facility(
        TEST_INTERNAL_FACILITY_1
    )
    assert sorted(result_capabilities) == sorted(expected_capabilities)


def test_get_facility_capabilities_no_input_facility():
    result_capabilities = ADAPTER.get_facility_capabilities_by_facility(None)
    assert result_capabilities == []


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_attributes_by_names"
)
def test_get_user_attributes_multiple_attributes(mock_request_1):
    test_user_attributes = get_entity_specific_attributes("user")
    attr_names = [attr.get("name") for attr in test_user_attributes]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_attributes_by_names = MagicMock(  # noqa E501
        return_value=test_user_attributes
    )

    expected_attributes = {
        attr.get("name"): attr.get("value") for attr in test_user_attributes
    }

    result_attributes = ADAPTER.get_user_attributes(TEST_USER, attr_names)
    assert result_attributes == expected_attributes


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_user_attributes_by_names"
)
def test_get_user_attributes_no_attributes(mock_request_1):
    # default attribute is loa
    default_attr = {
        "id": 1,
        "name": "perunUserAttribute_loa",
        "namespace": "urn:perun:user:attribute-def:virt",
        "display_name": "display name for loa",
        "type": "some type",
        "value": "sample value",
        "friendly_name": "loa",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_user_attributes_by_names = MagicMock(  # noqa E501
        return_value=[default_attr]
    )

    expected_attribute = {default_attr["name"]: default_attr["value"]}

    result_attributes = ADAPTER.get_user_attributes(TEST_USER, [])
    assert result_attributes == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_attributes_by_name"
)
@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_keys"
)
def test_get_entityless_attribute_valid_attribute(
    mock_request_1, mock_request_2
):
    test_entityless_attrs = get_entity_specific_attributes("entityless")
    test_entityless_attr_name = test_entityless_attrs[0].get("name")
    test_entityless_keys = ["entityless key 1", "entityless_key_2"]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_attributes_by_name = MagicMock(  # noqa E501
        return_value=test_entityless_attrs
    )

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_keys = MagicMock(  # noqa E501
        return_value=test_entityless_keys
    )

    expected_attrs = dict(zip(test_entityless_keys, test_entityless_attrs))
    result_attributes = ADAPTER.get_entityless_attribute(
        test_entityless_attr_name
    )
    assert result_attributes == expected_attrs


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_entityless_attributes_by_name"
)
def test_get_entityless_attribute_attr_id_missing(mock_request_1):
    test_entityless_attr_name = "perunEntitylessAttribute_orgAups"

    test_entityless_attr_without_id = {
        "id": None,
        "name": "perunEntitylessAttribute_orgAups",
        "namespace": "urn:perun:entityless:attribute-def:def:orgAups",
        "display_name": "display name for org aups",
        "type": "some type",
        "value": "sample value",
        "friendly_name": "orgAups",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_entityless_attributes_by_name = MagicMock(  # noqa E501
        return_value=[test_entityless_attr_without_id]
    )

    result_attributes = ADAPTER.get_entityless_attribute(
        test_entityless_attr_name
    )

    assert result_attributes == {}


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_vo_attributes_by_names"
)
def test_get_vo_attributes_single_attribute(mock_request_1):
    test_vo_attributes = get_entity_specific_attributes("vo")
    attr_names = [attr.get("name") for attr in test_vo_attributes]

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_vo_attributes_by_names = MagicMock(  # noqa E501
        return_value=test_vo_attributes
    )

    expected_attribute = {
        attr.get("name"): attr.get("value") for attr in test_vo_attributes
    }

    result_attribute = ADAPTER.get_vo_attributes(TEST_VO, attr_names)
    assert result_attribute == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_vo_attributes_by_names"
)
def test_get_vo_attributes_no_attributes(mock_request_1):
    # default attribute is id
    default_attr = {
        "id": 1,
        "name": "perunVoAttribute_id",
        "namespace": "urn:perun:vo:attribute-def:core",
        "display_name": "display name for loa",
        "type": "some type",
        "value": 10,
        "friendly_name": "id",
    }

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_vo_attributes_by_names = MagicMock(  # noqa E501
        return_value=[default_attr]
    )

    expected_attribute = {default_attr["name"]: default_attr["value"]}

    result_attribute = ADAPTER.get_vo_attributes(TEST_VO, [])
    assert result_attribute == expected_attribute


@patch(
    "perun.connector.perun_openapi.api.attributes_manager_api.AttributesManagerApi"
    ".get_attribute"
)
def test_get_facility_attribute_valid_attribute(mock_request_1):
    test_facility_attribute = get_entity_specific_attributes("facility")[0]
    test_attr_name = test_facility_attribute.get("name")

    perun_openapi.api.attributes_manager_api.AttributesManagerApi.get_attribute = MagicMock(  # noqa E501
        return_value=test_facility_attribute
    )

    expected_attribute = test_facility_attribute.get("value")
    result_attribute = ADAPTER.get_facility_attribute(
        TEST_INTERNAL_FACILITY_1, test_attr_name
    )

    assert result_attribute == expected_attribute


def test_get_attributes_multiple_attributes():
    test_perun_attrs = get_entity_specific_attributes("user")
    test_attr_names_map = {
        attr["namespace"] + ":" + attr["friendly_name"]: attr["name"]
        for attr in test_perun_attrs
    }
    processed_attributes = copy.deepcopy(test_perun_attrs)
    for attr in processed_attributes:
        attr.pop("namespace")

    expected_attributes = {
        attr.get("name"): attr for attr in processed_attributes
    }

    result_attributes = ADAPTER._get_attributes(
        test_perun_attrs, test_attr_names_map
    )
    assert result_attributes == expected_attributes


def test_get_attributes_empty_attributes():
    result_attributes = ADAPTER._get_attributes([], {})
    assert result_attributes == {}



from typing import List, Union, Optional

from perun.connector import perun_openapi
from perun.connector.adapters.AdapterInterface import AdapterInterface
from perun.connector.models.MemberStatusEnum import MemberStatusEnum
from perun.connector.utils.Logger import Logger
from perun.connector.models.Facility import Facility
from perun.connector.models.Group import Group
from perun.connector.models.Member import Member
from perun.connector.models.User import User
from perun.connector.models.UserExtSource import UserExtSource
from perun.connector.models.VO import VO
from perun.connector.perun_openapi import ApiClient, Configuration, ApiException
from perun.connector.perun_openapi.api.attributes_manager_api import AttributesManagerApi
from perun.connector.perun_openapi.api.facilities_manager_api import FacilitiesManagerApi
from perun.connector.perun_openapi.api.groups_manager_api import GroupsManagerApi
from perun.connector.perun_openapi.api.members_manager_api import MembersManagerApi
from perun.connector.perun_openapi.api.resources_manager_api import ResourcesManagerApi
from perun.connector.perun_openapi.api.searcher_api import SearcherApi
from perun.connector.perun_openapi.api.users_manager_api import UsersManagerApi
from perun.connector.perun_openapi.api.vos_manager_api import VosManagerApi
from perun.connector.perun_openapi.model.input_get_facilities import InputGetFacilities
from perun.connector.perun_openapi.model.input_set_user_ext_source_attributes import (
    InputSetUserExtSourceAttributes,
)
from perun.connector.utils.AttributeUtils import AttributeUtils


class PerunRpcAdapter(AdapterInterface):

    def __init__(self, config_data: dict[str, str], attrs_map):
        self._CONFIG = None
        self._logger = Logger.get_logger(self.__class__.__name__)
        self._BASIC_AUTH = "BasicAuth"
        self._BEARER_AUTH = "BearerAuth"
        self._API_KEY_AUTH = "ApiKeyAuth"

        self._set_up_openapi_config(config_data)

        self._RP_ID_ATTR = "perunFacilityAttr_rpID"
        self._ATTRIBUTE_UTILS = AttributeUtils(attrs_map)

    def _set_up_openapi_config(self, config_data: dict[str, str]) -> None:
        auth_type = config_data["auth_type"]
        self._CONFIG = Configuration(host=config_data["host"])

        if auth_type == self._BASIC_AUTH:
            self._CONFIG.username = config_data["username"]
            self._CONFIG.password = config_data["password"]
        elif auth_type == self._BEARER_AUTH:
            self._CONFIG.access_token = config_data["access_token"]
        elif auth_type == self._API_KEY_AUTH:
            self._CONFIG.api_key[self._API_KEY_AUTH] = config_data["api_key"]
        else:
            exception_message = (
                f'Authentication type "{auth_type}" is not a '
                f"supported way of authentication, please "
                f'set "auth_type" to one of "'
                f'{self._BASIC_AUTH}", "{self._BEARER_AUTH}" '
                f'or "{self._API_KEY_AUTH}"'
            )
            raise ValueError(exception_message)

    def get_perun_user(self, idp_id: str, uids: List[str]) -> Optional[User]:
        with ApiClient(self._CONFIG) as api_client:
            api_instance = UsersManagerApi(api_client)
            for uid in uids:
                try:
                    user = \
                        api_instance.get_user_by_ext_source_name_and_ext_login(
                            ext_login=uid, ext_source_name=idp_id
                        )
                    name = ""
                    for user_attr in [
                        "title_before",
                        "first_name",
                        "middle_name",
                        "last_name",
                        "title_after",
                    ]:
                        if user[user_attr] is not None:
                            name += user[user_attr] + " "

                    return User(user["id"], name.strip())
                except ApiException as ex:
                    if '"name":"UserExtSourceNotExistsException"' in ex.body:
                        continue
                    raise ex
            return None

    def _get_group_unique_name(
            self,
            attributes_api_instance: AttributesManagerApi,
            group_name: str,
            group_id: int,
    ) -> str:
        attr = attributes_api_instance.get_attribute(
            group=group_id,
            attribute_name="urn:perun:group:attribute-def:virt:voShortName",
        )
        return f'{attr["value"]}:{group_name}'

    def _create_internal_representation_groups(self,
                                               input_groups: List[
                                                   perun_openapi.model.group.Group
                                               ],
                                               converted_groups: List[Group],
                                               attributes_api_instance:
                                               AttributesManagerApi) -> None:
        unique_ids = []
        for group in input_groups:
            if group["id"] not in unique_ids:
                group["unique_name"] = self._get_group_unique_name(
                    attributes_api_instance, group["name"], group["id"]
                )
                converted_groups.append(
                    Group(
                        group["id"],
                        self.get_vo(vo_id=group["vo_id"]),
                        group["uuid"],
                        group["name"],
                        group["unique_name"],
                        group["description"],
                    )
                )
                unique_ids.append(group["id"])

    def get_member_groups(self, user: Union[User, int], vo: Union[VO, int]) -> List[Group]:
        with ApiClient(self._CONFIG) as api_client:
            members_api_instance = MembersManagerApi(api_client)
            groups_api_instance = GroupsManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            converted_groups = []
            vo_id = AdapterInterface.get_object_id(vo)
            user_id = AdapterInterface.get_object_id(user)
            try:
                member = members_api_instance.get_member_by_user(
                    vo_id, user_id
                )
                member_groups = []
                if member:
                    member_groups = groups_api_instance.get_all_member_groups(
                        member["id"]
                    )
                self._create_internal_representation_groups(member_groups,
                                                            converted_groups,
                                                            attributes_api_instance)  # noqa E501
            except ApiException as e:
                self._logger.warning(f' OpenAPI raised an exception: "{e}"')

            return converted_groups

    def get_sp_groups_by_facility(self, facility: Union[Facility, int]) -> List[Group]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            facilities_api_instance = FacilitiesManagerApi(api_client)
            resources_api_instance = ResourcesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            resources = (
                facilities_api_instance.get_assigned_resources_for_facility(
                    facility_id
                )
            )

            resources_ids = [resource.id for resource in resources]

            sp_groups = []
            for resource_id in resources_ids:
                groups = resources_api_instance.get_assigned_groups(
                    resource_id
                )

                self._create_internal_representation_groups(groups,
                                                            sp_groups,
                                                            attributes_api_instance)  # noqa E501
            return sp_groups

    def get_sp_groups_by_rp_id(self, rp_id: str) -> List[Group]:
        facility = self.get_facility_by_rp_identifier(rp_id)
        return self.get_sp_groups_by_facility(facility)

    def get_group_by_name(self, vo: Union[VO, int], name: str) -> Group:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)
            groups_api_instance = GroupsManagerApi(api_client)

            vo_id = AdapterInterface.get_object_id(vo)
            group = groups_api_instance.get_group_by_name(vo_id, name)
            group_external_representation = [group]
            converted_group = []
            self._create_internal_representation_groups(
                group_external_representation,
                converted_group,
                attributes_api_instance)
            return converted_group[0]

    def get_vo(self, short_name=None, vo_id=None) -> Optional[VO]:
        with ApiClient(self._CONFIG) as api_client:
            vos_api_instance = VosManagerApi(api_client)

            if short_name and vo_id:
                raise ValueError(
                    "VO can be obtained either by its short_name or id, "
                    "not both "
                    "at the same time."
                )
            elif vo_id:
                vo_lookup_method = vos_api_instance.get_vo_by_id
                vo_lookup_attribute = vo_id
                identifier = "id"
            elif short_name:
                vo_lookup_method = vos_api_instance.get_vo_by_short_name
                vo_lookup_attribute = short_name
                identifier = "short name"
            else:
                raise ValueError(
                    "Neither short_name nor id was provided, please specify "
                    "exactly one to find VO by."
                )

            try:
                vo = vo_lookup_method(vo_lookup_attribute)
                return VO(vo.id, vo.name, vo.short_name)
            except ApiException as ex:
                vo_not_found = '"name":"VoNotExistsException"' in ex.body

                if vo_not_found:
                    self._logger.warning(
                        f'VO looked up by {identifier} "'
                        f'{vo_lookup_attribute}" does not exist in Perun.'
                    )
                    return None
                raise ex

    def get_facility_by_rp_identifier(
            self,
            rp_identifier: str,
    ) -> Optional[Facility]:
        with ApiClient(self._CONFIG) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)

            attr_name = self._ATTRIBUTE_UTILS.get_rpc_attr_name(
                self._RP_ID_ATTR
            )

            facilities = facilities_api_instance.get_facilities_by_attribute(
                attribute_name=attr_name, attribute_value=rp_identifier
            )

            if not facilities:
                self._logger.warning(
                    f"No facility with rpID '{rp_identifier}' found."
                )
                return None

            if len(facilities) > 1:
                self._logger.warning(
                    f"There is more than one facility with rpID '"
                    f"{rp_identifier}'."
                )
                return None
            return Facility(
                facilities[0]["id"],
                facilities[0]["name"],
                facilities[0]["description"],
                rp_identifier,
            )

    def get_users_groups_on_facility(
            self, facility: Union[Facility, int], user: Union[User, int]
    ) -> List[Group]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG) as api_client:
            users_api_instance = UsersManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            user_id = AdapterInterface.get_object_id(user)
            users_groups_on_facility = (
                users_api_instance.get_groups_for_facility_where_user_is_active(  # noqa E501
                    user_id,
                    facility_id,
                )
            )
            converted_groups = []
            self._create_internal_representation_groups(
                users_groups_on_facility, converted_groups,
                attributes_api_instance)
            return converted_groups

    def get_users_groups_on_facility_by_rp_id(
            self, rp_identifier: str, user: Union[User, int]
    ) -> List[Group]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_users_groups_on_facility(facility, user)

    def _get_rp_id(self, facility: Facility) -> str:
        return self.get_facility_attributes(facility, [self._RP_ID_ATTR]) \
            .get(self._RP_ID_ATTR)

    # TODO test this method once SearcherAPI is supported on Devel
    def get_facilities_by_attribute_value(
            self, attribute: dict[str, str]
    ) -> List[Facility]:
        if len(attribute) != 1:
            self._logger.warning(
                f'Attribute must contain exactly one name and one value. '
                f'Given attribute contains: "{attribute}".'
            )
            return []

        with ApiClient(self._CONFIG) as api_client:
            searcher_api = SearcherApi(api_client)

            attribute_to_match_in_facilities = InputGetFacilities(attribute)
            perun_facilities = searcher_api.get_facilities(
                attribute_to_match_in_facilities
            )

            facilities = []
            for perun_facility in perun_facilities:
                facility = Facility(perun_facility['id'],
                                    perun_facility['name'],
                                    perun_facility['description'],
                                    "")
                facility.rp_id = self._get_rp_id(facility)
                facilities.append(facility)

        return facilities

    def get_facility_attributes(
            self, facility: Union[Facility, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)

            attr_names_map = self._ATTRIBUTE_UTILS.get_rpc_attr_names(
                attr_names
            )
            perun_attrs = (
                attributes_api_instance.get_facility_attributes_by_names(
                    facility_id, list(attr_names_map.keys())
                )
            )
            facility_attrs = self._get_attributes(perun_attrs, attr_names_map)
            return {
                facility_attr_name: facility_attr["value"]
                for facility_attr_name, facility_attr in facility_attrs.items()
            }

    def get_user_ext_source(
            self, ext_source_name: str, ext_source_login: str
    ) -> UserExtSource:
        with ApiClient(self._CONFIG) as api_client:
            users_api_instance = UsersManagerApi(api_client)

            user_ext_source_perun = \
                users_api_instance.get_user_ext_source_by_ext_login_and_ext_source_name(  # noqa E501
                    ext_source_name=ext_source_name,
                    ext_source_login=ext_source_login
                )

            ext_source_id = user_ext_source_perun["id"]
            login = user_ext_source_perun["login"]

            ext_source_details = user_ext_source_perun["ext_source"]
            name = ext_source_details["name"]

            user = self.get_perun_user(ext_source_name, [ext_source_login])

            return UserExtSource(ext_source_id, name, login, user)

    def update_user_ext_source_last_access(
            self, user_ext_source: Union[UserExtSource, int]
    ) -> None:
        user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)

        with ApiClient(self._CONFIG) as api_client:
            users_api_instance = UsersManagerApi(api_client)

            users_api_instance.update_user_ext_source_last_access(
                user_ext_source_id
            )

    def get_user_ext_source_attributes(
            self, user_ext_source: Union[UserExtSource, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)

            attr_names_map = self._ATTRIBUTE_UTILS.get_rpc_attr_names(
                attr_names
            )
            perun_attrs = \
                attributes_api_instance.get_user_ext_source_attributes_by_names(  # noqa E501
                    user_ext_source=user_ext_source_id,
                    attr_names=list(attr_names_map.keys()),
                )
            return self._get_attributes(perun_attrs, attr_names_map)

    def set_user_ext_source_attributes(
            self,
            user_ext_source: Union[UserExtSource, int],
            attributes: List[
                dict[
                    str, Union[
                        str, Optional[int], bool, List[str], dict[str, str]]
                ]
            ],
    ) -> None:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            user_ext_source_id = AdapterInterface.get_object_id(user_ext_source)
            attributes_api_instance.set_user_ext_source_attributes(
                InputSetUserExtSourceAttributes(user_ext_source_id, attributes)
            )

    def get_member_status_by_user_and_vo(
            self, user: Union[User, int], vo: Union[VO, int]
    ) -> Optional[str]:

        member = self.get_member_by_user(user, vo)

        if member is not None:
            return member.status

        return None

    def is_user_in_vo_by_short_name(self, user: Union[User, int], vo_short_name: str) -> bool:
        user_id = AdapterInterface.get_object_id(user)
        if not user_id:
            raise ValueError("User's ID is empty")

        if not vo_short_name:
            raise ValueError("VO short name is empty")

        vo_of_user = self.get_vo(short_name=vo_short_name)

        if vo_of_user is None:
            self._logger.debug(
                f'No VO with short name "{vo_short_name}" found')
            return False

        user_status = self.get_member_status_by_user_and_vo(user, vo_of_user)
        valid_status = MemberStatusEnum.VALID

        return user_status == valid_status

    def get_member_by_user(self, user: Union[User, int], vo: Union[VO, int]) -> Optional[Member]:
        with ApiClient(self._CONFIG) as api_client:
            members_api_instance = MembersManagerApi(api_client)

            user_id = AdapterInterface.get_object_id(user)
            vo_id = AdapterInterface.get_object_id(vo)

            try:
                member = members_api_instance.get_member_by_user(vo_id,
                                                                 user_id)
                return Member(member["id"], vo, member["status"])
            except ApiException as ex:
                user_not_found = '"name":"UserNotExistsException"' in ex.body
                vo_not_found = '"name":"VoNotExistsException"' in ex.body
                member_not_exists = '"name":"MemberNotExistsException"' in \
                                    ex.body

                if user_not_found:
                    self._logger.warning(
                        f'User with id "{user_id}" does not exist in '
                        f'Perun.'
                    )
                if vo_not_found:
                    self._logger.warning(
                        f'VO with id "{vo_id}" does not '
                        f'exist in Perun.'
                    )
                if member_not_exists:
                    self._logger.warning(
                        f'Member with VO "{vo_id}" and user id "'
                        f'{user_id}" does not exist in Perun.'
                    )

                if user_not_found or vo_not_found or member_not_exists:
                    return None

                raise ex

    def get_resource_capabilities_by_facility(
            self, facility: Union[Facility, int], user_groups: List[Union[Group, int]]
    ) -> List[str]:
        capabilities = []
        if facility is None:
            return capabilities

        with ApiClient(self._CONFIG) as api_client:
            facilities_api_instance = FacilitiesManagerApi(api_client)
            resources_api_instance = ResourcesManagerApi(api_client)
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)
            resources = (
                facilities_api_instance.get_assigned_resources_for_facility(
                    facility_id
                )
            )
            user_groups_ids = [AdapterInterface.get_object_id(user_group) for user_group in user_groups]
            for resource in resources:
                resource_groups = resources_api_instance.get_assigned_groups(
                    resource["id"]
                )

                resource_capabilities = attributes_api_instance.get_attribute(
                    resource=resource["id"],
                    attribute_name="urn:perun:resource:attribute-def:def"
                                   ":capabilities"
                )["value"]

                if resource_capabilities is None:
                    continue

                for resource_group in resource_groups:
                    if resource_group["id"] in user_groups_ids:
                        capabilities.extend(resource_capabilities)
                        break
        return capabilities

    def get_resource_capabilities_by_rp_id(
            self, rp_identifier: str, user_groups: List[Union[Group, int]]
    ) -> List[str]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_resource_capabilities_by_facility(facility, user_groups)

    def get_facility_capabilities_by_facility(self, facility: Union[Facility, int]) -> List[str]:
        if facility is None:
            return []

        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)

            facility_capabilities = attributes_api_instance.get_attribute(
                facility=facility_id,
                attribute_name="urn:perun:facility:attribute-def:def"
                               ":capabilities",
            )["value"]

            return facility_capabilities

    def get_facility_capabilities_by_rp_id(self, rp_identifier: str) -> List[str]:
        facility = self.get_facility_by_rp_identifier(rp_identifier)
        return self.get_facility_capabilities_by_facility(facility)

    def get_user_attributes(
            self, user: Union[User, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        default_attribute_name = "perunUserAttribute_loa"
        if not attr_names:
            attr_names.append(default_attribute_name)

        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            user_id = AdapterInterface.get_object_id(user)

            attr_names_map = self._ATTRIBUTE_UTILS.get_rpc_attr_names(
                attr_names
            )

            perun_attrs = attributes_api_instance.get_user_attributes_by_names(
                user_id, list(attr_names_map.keys())
            )

            user_attrs = self._get_attributes(perun_attrs, attr_names_map)

            return {
                user_attr_name: user_attr["value"]
                for user_attr_name, user_attr in user_attrs.items()
            }

    def get_entityless_attribute(
            self, attr_name: str
    ) -> Union[str, Optional[int], bool, List[str], dict[str, str]]:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            attributes = {}
            perun_attr_values = (
                attributes_api_instance.get_entityless_attributes_by_name(
                    attr_name=self._ATTRIBUTE_UTILS.get_rpc_attr_name(
                        attr_name)
                )
            )

            attr_id = perun_attr_values[0].get("id")
            if attr_id is None:
                return attributes

            perun_attr_keys = attributes_api_instance.get_entityless_keys(
                attr_id
            )

            return dict(zip(perun_attr_keys, perun_attr_values))

    def get_vo_attributes(
            self, vo: Union[VO, int], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        default_attribute_name = "perunVoAttribute_id"
        if not attr_names:
            attr_names.append(default_attribute_name)

        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            vo_id = AdapterInterface.get_object_id(vo)

            attr_names_map = self._ATTRIBUTE_UTILS.get_rpc_attr_names(
                attr_names
            )

            perun_attrs = attributes_api_instance.get_vo_attributes_by_names(
                vo_id, list(attr_names_map.keys())
            )

            vo_attrs = self._get_attributes(perun_attrs, attr_names_map)

            return {
                vo_attr_name: user_attr["value"]
                for vo_attr_name, user_attr in vo_attrs.items()
            }

    def get_facility_attribute(
            self, facility: Union[Facility, int], attr_name: str
    ) -> Union[str, Optional[int], bool, List[str], dict[str, str]]:
        with ApiClient(self._CONFIG) as api_client:
            attributes_api_instance = AttributesManagerApi(api_client)

            facility_id = AdapterInterface.get_object_id(facility)

            attr_name = self._ATTRIBUTE_UTILS.get_rpc_attr_name(attr_name)
            perun_attr = attributes_api_instance.get_attribute(
                facility=facility_id, attribute_name=attr_name
            )

            return perun_attr["value"]

    def _get_attributes(
            self, perun_attrs: List[dict[str, str]],
            attr_names_map: dict[str, str]
    ) -> dict[
        str,
        dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]],
    ]:
        attributes = {}
        for perun_attr in perun_attrs:
            perun_attr_name = perun_attr["namespace"] + ":" + perun_attr[
                "friendly_name"]

            attributes[attr_names_map[perun_attr_name]] = {
                "id": perun_attr["id"],
                "name": attr_names_map[perun_attr_name],
                "display_name": perun_attr["display_name"],
                "type": perun_attr["type"],
                "value": perun_attr["value"],
                "friendly_name": perun_attr["friendly_name"]
            }

        return attributes




from perun.connector.connectors.LdapConnector import LdapConnector
from unittest.mock import patch, MagicMock

import pytest

from perun.connector.utils.ConfigStore import ConfigStore

loaded_config = ConfigStore.get_ldapc_config()
CONNECTOR = LdapConnector(loaded_config)

TEST_DATA = {
    "perunUserId": 1,
    "displayName": "Foe Toe",
    "cn": "Foe Toe",
    "memberOf": ['perunGroupId=1,perunVoId=1,dc=perun,dc=cesnet,dc=cz',
                 'perunGroupId=2,perunVoId=1,dc=perun,dc=cesnet,dc=cz']
}
TEST_DATA_2 = {
    "perunUserId": 2,
    "displayName": "Joe Doe",
    "cn": "Joe Doe",
    "memberOf": ['perunGroupId=1,perunVoId=1,dc=perun,dc=cesnet,dc=cz',
                 'perunGroupId=2,perunVoId=1,dc=perun,dc=cesnet,dc=cz']
}

TEST_ENTRIES = [TEST_DATA, TEST_DATA_2]
TEST_ENTRY = [TEST_DATA]
BASE = "ldap.base"
FILTERS = "(name=test_filters)"


@patch(
    "perun.connector.connectors.LdapConnector.LdapConnector._search"
)
def test_search_for_entity_not_found(mock_request):
    CONNECTOR._search = MagicMock(
        return_value=None
    )
    result = CONNECTOR.search_for_entity(BASE, FILTERS)
    assert not result


@patch(
    "perun.connector.connectors.LdapConnector.LdapConnector._search"
)
def test_search_for_entity_found(mock_request):
    CONNECTOR._search = MagicMock(
        return_value=TEST_ENTRY
    )
    result = CONNECTOR.search_for_entity(BASE, FILTERS)
    assert result == TEST_DATA


@patch(
    "perun.connector.connectors.LdapConnector.LdapConnector._search"
)
def test_search_for_entity_found_invalid(mock_request):
    CONNECTOR._search = MagicMock(
        return_value=TEST_ENTRIES
    )
    expected_error_message = 'ldap_connector.search_for_entity - ' \
                             'More than one entity found.' + ' query base:' \
                             + BASE + ', filter: ' + FILTERS + '.' + \
                             ' Hint: Use method ''search_for_entities if ' \
                             'you expect array of entities.'

    with pytest.raises(Exception) as error:
        _ = CONNECTOR.search_for_entity(BASE, FILTERS)

    assert str(error.value.args[0]) == expected_error_message


@patch(
    "perun.connector.connectors.LdapConnector.LdapConnector._search"
)
def test_search_for_entities_not_found(mock_request):
    CONNECTOR._search = MagicMock(
        return_value=None
    )
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert not result


@patch(
    "perun.connector.connectors.LdapConnector.LdapConnector._search"
)
def test_search_for_entities_found(mock_request):
    CONNECTOR._search = MagicMock(
        return_value=TEST_ENTRIES
    )
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert result == TEST_ENTRIES

    CONNECTOR._search = MagicMock(
        return_value=TEST_ENTRY
    )
    result = CONNECTOR.search_for_entities(BASE, FILTERS)
    assert result == TEST_ENTRY



import inspect
from typing import List, Union, Optional

from perun.connector.adapters.PerunRpcAdapter import PerunRpcAdapter
from perun.connector.adapters.LdapAdapter import LdapAdapter
from perun.connector.adapters.LdapAdapter import AdapterSkipException
from perun.connector.utils.Logger import Logger

from perun.connector.adapters.AdapterInterface import AdapterInterface
from perun.connector.models.Facility import Facility
from perun.connector.models.Group import Group
from perun.connector.models.User import User
from perun.connector.models.UserExtSource import UserExtSource
from perun.connector.models.VO import VO
from perun.connector.perun_openapi import ApiException


class AdaptersManager(AdapterInterface):
    def __init__(self, config, attrs_map):
        self._logger = Logger.get_logger(self.__class__.__name__)
        self._STARTING_PRIORITY = 1
        self.adapters = {}

        adapters_info = config["adapters"]

        for adapter_info in adapters_info:
            config_data = adapter_info.copy()
            adapter_type = config_data.pop("type")
            priority = config_data.pop("priority")

            if adapter_type == "ldap":
                ldap_adapter = LdapAdapter(config_data, attrs_map)

                self.adapters[priority] = {
                    "name": "ldap_adapter",
                    "adapter": ldap_adapter,
                }
            elif adapter_type == "openApi":
                rpc_adapter = PerunRpcAdapter(config_data, attrs_map)

                self.adapters[priority] = {
                    "name": "rpc_adapter",
                    "adapter": rpc_adapter,
                }
            else:
                self._logger.warning(
                    f'Config file includes unsupported adapter type "'
                    f'{adapter_type}"'
                )

    def _execute_method_by_priority(self, method_name: str, *args):
        current_priority = self._STARTING_PRIORITY
        current_adapter = self.adapters.get(current_priority)

        while current_adapter is not None:
            adapter_impl = current_adapter["adapter"]
            try:
                return getattr(adapter_impl, method_name)(*args)
            except AdapterSkipException:
                self._logger.warning(
                    f'Method "{method_name}" is not supported by '
                    f'{current_adapter["name"]}. Going to try another '
                    f'adapter if available.')
                current_priority += 1
                current_adapter = self.adapters.get(current_priority)
            except ApiException as ex:
                if 'notexistsexception"' in ex.body.lower():
                    self._logger.warning(
                        "Requested entity doesn't exist in Perun")
                raise
            except Exception as ex:
                self._logger.warning(
                    f'Method "{method_name}" could not be executed '
                    f'successfully by {current_adapter["name"]}, exception '
                    f'occurred: "{ex}"')
                raise

        raise Exception(
            f'None of the provided adapters was able to resolve method "'
            f'{method_name}"'
        )

    def _get_caller_name(self):
        return inspect.stack()[1].function

    def get_perun_user(self, idp_id: str, uids: List[str]) -> Optional[User]:
        return self._execute_method_by_priority(
            self._get_caller_name(), idp_id, uids
        )

    def get_group_by_name(self, vo: Union[int, VO], name: str) -> Group:
        return self._execute_method_by_priority(
            self._get_caller_name(), vo, name
        )

    def get_vo(self, short_name: str, vo_id: int) -> VO:
        return self._execute_method_by_priority(
            self._get_caller_name(), short_name, vo_id
        )

    def get_member_groups(self, user: Union[int, User], vo: Union[int, VO]) -> List[Group]:
        return self._execute_method_by_priority(
            self._get_caller_name(), user, vo
        )

    def get_sp_groups_by_facility(self, facility: Union[Facility, int]) -> List[Group]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility
        )

    def get_sp_groups_by_rp_id(self, rp_id: str) -> List[Group]:
        return self._execute_method_by_priority(
            self._get_caller_name(), rp_id
        )

    def get_user_attributes(
            self, user: Union[int, User], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), user, attr_names
        )

    def get_entityless_attribute(
            self, attr_name: str
    ) -> Union[str, Optional[int], bool, List[str], dict[str, str]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), attr_name
        )

    def get_vo_attributes(
            self, vo: Union[int, VO], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), vo, attr_names
        )

    def get_facility_attribute(
            self, facility: Union[int, Facility], attr_name: str
    ) -> Union[str, Optional[int], bool, List[str], dict[str, str]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility, attr_name
        )

    def get_facility_by_rp_identifier(
            self, rp_identifier: str
    ) -> Facility:
        return self._execute_method_by_priority(
            self._get_caller_name(), rp_identifier
        )

    def get_users_groups_on_facility_by_rp_id(
            self, rp_identifier: str, user: Union[User, int]
    ) -> List[Group]:
        return self._execute_method_by_priority(
            self._get_caller_name(), rp_identifier, user
        )

    def get_users_groups_on_facility(
            self, facility: Union[Facility, int], user: Union[User, int]
    ) -> List[Group]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility, user
        )

    def get_facilities_by_attribute_value(
            self, attribute: dict[str, str]
    ) -> List[Facility]:
        return self._execute_method_by_priority(
            self._get_caller_name(), attribute
        )

    def get_facility_attributes(
            self, facility: Union[int, Facility], attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility, attr_names
        )

    def get_user_ext_source(
            self, ext_source_name: str, ext_source_login: str
    ) -> UserExtSource:
        return self._execute_method_by_priority(
            self._get_caller_name(), ext_source_name, ext_source_login
        )

    def update_user_ext_source_last_access(self, user_ext_source: str) -> None:
        return self._execute_method_by_priority(
            self._get_caller_name(), user_ext_source
        )

    def get_user_ext_source_attributes(
            self, user_ext_source: Union[int, UserExtSource],
            attr_names: List[str]
    ) -> dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]:
        return self._execute_method_by_priority(
            self._get_caller_name(), user_ext_source
        )

    def set_user_ext_source_attributes(
            self, user_ext_source: Union[int, UserExtSource],
            attributes: List[
                dict[str, Union[str, Optional[int], bool, List[str], dict[str, str]]]]
    ) -> None:
        return self._execute_method_by_priority(
            self._get_caller_name(), user_ext_source
        )

    def get_member_status_by_user_and_vo(self, user: Union[int, User], vo: Union[int, VO]) -> str:
        return self._execute_method_by_priority(
            self._get_caller_name(), user, vo
        )

    def is_user_in_vo_by_short_name(self, user: Union[int, User], vo_short_name: str) -> bool:
        return self._execute_method_by_priority(
            self._get_caller_name(), user, vo_short_name
        )

    def get_resource_capabilities_by_facility(
            self, facility: Union[Facility, int], user_groups: List[Union[Group, int]]
    ) -> List[str]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility, user_groups
        )

    def get_resource_capabilities_by_rp_id(
            self, rp_identifier: str, user_groups: List[Union[Group, int]]
    ) -> List[str]:
        return self._execute_method_by_priority(
            self._get_caller_name(), rp_identifier, user_groups
        )

    def get_facility_capabilities_by_rp_id(self, rp_identifier: str) -> List[str]:
        return self._execute_method_by_priority(
            self._get_caller_name(), rp_identifier
        )

    def get_facility_capabilities_by_facility(self, facility: Union[Facility, int]) -> List[str]:
        return self._execute_method_by_priority(
            self._get_caller_name(), facility
        )



# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from perun.connector.perun_openapi.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from perun.connector.perun_openapi.model.action_type import ActionType
from perun.connector.perun_openapi.model.add_user_ext_source_input import AddUserExtSourceInput
from perun.connector.perun_openapi.model.app_state import AppState
from perun.connector.perun_openapi.model.app_type import AppType
from perun.connector.perun_openapi.model.application import Application
from perun.connector.perun_openapi.model.application_form import ApplicationForm
from perun.connector.perun_openapi.model.application_form_item import ApplicationFormItem
from perun.connector.perun_openapi.model.application_form_item_data import ApplicationFormItemData
from perun.connector.perun_openapi.model.application_mail import ApplicationMail
from perun.connector.perun_openapi.model.applications_order_column import ApplicationsOrderColumn
from perun.connector.perun_openapi.model.applications_page_query import ApplicationsPageQuery
from perun.connector.perun_openapi.model.assigned_group import AssignedGroup
from perun.connector.perun_openapi.model.assigned_member import AssignedMember
from perun.connector.perun_openapi.model.assigned_resource import AssignedResource
from perun.connector.perun_openapi.model.attribute import Attribute
from perun.connector.perun_openapi.model.attribute_action import AttributeAction
from perun.connector.perun_openapi.model.attribute_all_of import AttributeAllOf
from perun.connector.perun_openapi.model.attribute_definition import AttributeDefinition
from perun.connector.perun_openapi.model.attribute_definition_all_of import AttributeDefinitionAllOf
from perun.connector.perun_openapi.model.attribute_policy import AttributePolicy
from perun.connector.perun_openapi.model.attribute_policy_collection import AttributePolicyCollection
from perun.connector.perun_openapi.model.attribute_rights import AttributeRights
from perun.connector.perun_openapi.model.audit_event import AuditEvent
from perun.connector.perun_openapi.model.audit_message import AuditMessage
from perun.connector.perun_openapi.model.audit_messages_page_query import AuditMessagesPageQuery
from perun.connector.perun_openapi.model.auditable import Auditable
from perun.connector.perun_openapi.model.auditable_all_of import AuditableAllOf
from perun.connector.perun_openapi.model.author import Author
from perun.connector.perun_openapi.model.author_all_of import AuthorAllOf
from perun.connector.perun_openapi.model.authorship import Authorship
from perun.connector.perun_openapi.model.authorship_all_of import AuthorshipAllOf
from perun.connector.perun_openapi.model.authz_roles import AuthzRoles
from perun.connector.perun_openapi.model.ban import Ban
from perun.connector.perun_openapi.model.ban_all_of import BanAllOf
from perun.connector.perun_openapi.model.ban_on_facility import BanOnFacility
from perun.connector.perun_openapi.model.ban_on_facility_all_of import BanOnFacilityAllOf
from perun.connector.perun_openapi.model.ban_on_resource import BanOnResource
from perun.connector.perun_openapi.model.ban_on_resource_all_of import BanOnResourceAllOf
from perun.connector.perun_openapi.model.ban_on_vo import BanOnVo
from perun.connector.perun_openapi.model.ban_on_vo_all_of import BanOnVoAllOf
from perun.connector.perun_openapi.model.brand import Brand
from perun.connector.perun_openapi.model.candidate import Candidate
from perun.connector.perun_openapi.model.candidate_all_of import CandidateAllOf
from perun.connector.perun_openapi.model.category import Category
from perun.connector.perun_openapi.model.category_all_of import CategoryAllOf
from perun.connector.perun_openapi.model.consent import Consent
from perun.connector.perun_openapi.model.consent_all_of import ConsentAllOf
from perun.connector.perun_openapi.model.consent_hub import ConsentHub
from perun.connector.perun_openapi.model.consent_hub_all_of import ConsentHubAllOf
from perun.connector.perun_openapi.model.consent_status import ConsentStatus
from perun.connector.perun_openapi.model.destination import Destination
from perun.connector.perun_openapi.model.destination_all_of import DestinationAllOf
from perun.connector.perun_openapi.model.destination_propagation_type import DestinationPropagationType
from perun.connector.perun_openapi.model.destination_type import DestinationType
from perun.connector.perun_openapi.model.enriched_facility import EnrichedFacility
from perun.connector.perun_openapi.model.enriched_group import EnrichedGroup
from perun.connector.perun_openapi.model.enriched_host import EnrichedHost
from perun.connector.perun_openapi.model.enriched_resource import EnrichedResource
from perun.connector.perun_openapi.model.enriched_vo import EnrichedVo
from perun.connector.perun_openapi.model.entityless_attributes_by_keys import EntitylessAttributesByKeys
from perun.connector.perun_openapi.model.ext_source import ExtSource
from perun.connector.perun_openapi.model.ext_source_all_of import ExtSourceAllOf
from perun.connector.perun_openapi.model.ext_source_object import ExtSourceObject
from perun.connector.perun_openapi.model.facility import Facility
from perun.connector.perun_openapi.model.facility_all_of import FacilityAllOf
from perun.connector.perun_openapi.model.facility_propagation_state import FacilityPropagationState
from perun.connector.perun_openapi.model.facility_state import FacilityState
from perun.connector.perun_openapi.model.gen_data_node import GenDataNode
from perun.connector.perun_openapi.model.gen_member_data_node import GenMemberDataNode
from perun.connector.perun_openapi.model.group import Group
from perun.connector.perun_openapi.model.group_all_of import GroupAllOf
from perun.connector.perun_openapi.model.group_member_data import GroupMemberData
from perun.connector.perun_openapi.model.group_member_relation import GroupMemberRelation
from perun.connector.perun_openapi.model.group_resource_status import GroupResourceStatus
from perun.connector.perun_openapi.model.groups_order_column import GroupsOrderColumn
from perun.connector.perun_openapi.model.groups_page_query import GroupsPageQuery
from perun.connector.perun_openapi.model.hashed_gen_data import HashedGenData
from perun.connector.perun_openapi.model.host import Host
from perun.connector.perun_openapi.model.host_all_of import HostAllOf
from perun.connector.perun_openapi.model.identity import Identity
from perun.connector.perun_openapi.model.inline_response200 import InlineResponse200
from perun.connector.perun_openapi.model.input_add_application_mail_for_group import InputAddApplicationMailForGroup
from perun.connector.perun_openapi.model.input_add_application_mail_for_vo import InputAddApplicationMailForVo
from perun.connector.perun_openapi.model.input_add_destination_to_multiple_services import InputAddDestinationToMultipleServices
from perun.connector.perun_openapi.model.input_add_destinations_defined_by_hosts_on_facility import InputAddDestinationsDefinedByHostsOnFacility
from perun.connector.perun_openapi.model.input_assign_resource_tag_to_resource import InputAssignResourceTagToResource
from perun.connector.perun_openapi.model.input_attribute_definition import InputAttributeDefinition
from perun.connector.perun_openapi.model.input_attribute_policy_collections import InputAttributePolicyCollections
from perun.connector.perun_openapi.model.input_attribute_rights import InputAttributeRights
from perun.connector.perun_openapi.model.input_copy_resource import InputCopyResource
from perun.connector.perun_openapi.model.input_create_attribute_definition import InputCreateAttributeDefinition
from perun.connector.perun_openapi.model.input_create_authorship import InputCreateAuthorship
from perun.connector.perun_openapi.model.input_create_category import InputCreateCategory
from perun.connector.perun_openapi.model.input_create_category1 import InputCreateCategory1
from perun.connector.perun_openapi.model.input_create_member_for_candidate import InputCreateMemberForCandidate
from perun.connector.perun_openapi.model.input_create_member_for_user import InputCreateMemberForUser
from perun.connector.perun_openapi.model.input_create_member_from_ext_source import InputCreateMemberFromExtSource
from perun.connector.perun_openapi.model.input_create_owner import InputCreateOwner
from perun.connector.perun_openapi.model.input_create_publication import InputCreatePublication
from perun.connector.perun_openapi.model.input_create_publication_system import InputCreatePublicationSystem
from perun.connector.perun_openapi.model.input_create_resource_tag_with_resource_tag import InputCreateResourceTagWithResourceTag
from perun.connector.perun_openapi.model.input_create_service import InputCreateService
from perun.connector.perun_openapi.model.input_create_service_user import InputCreateServiceUser
from perun.connector.perun_openapi.model.input_create_services_package import InputCreateServicesPackage
from perun.connector.perun_openapi.model.input_create_sponsored_member import InputCreateSponsoredMember
from perun.connector.perun_openapi.model.input_create_sponsored_member1 import InputCreateSponsoredMember1
from perun.connector.perun_openapi.model.input_create_sponsored_member_from_csv import InputCreateSponsoredMemberFromCSV
from perun.connector.perun_openapi.model.input_create_thanks import InputCreateThanks
from perun.connector.perun_openapi.model.input_create_vo_with_vo import InputCreateVoWithVo
from perun.connector.perun_openapi.model.input_delete_groups import InputDeleteGroups
from perun.connector.perun_openapi.model.input_delete_resource_tag import InputDeleteResourceTag
from perun.connector.perun_openapi.model.input_entityless_attribute import InputEntitylessAttribute
from perun.connector.perun_openapi.model.input_form_item_data import InputFormItemData
from perun.connector.perun_openapi.model.input_form_item_data1 import InputFormItemData1
from perun.connector.perun_openapi.model.input_get_all_resources_by_resource_tag import InputGetAllResourcesByResourceTag
from perun.connector.perun_openapi.model.input_get_facilities import InputGetFacilities
from perun.connector.perun_openapi.model.input_get_members_by_user_attributes import InputGetMembersByUserAttributes
from perun.connector.perun_openapi.model.input_get_messages_page import InputGetMessagesPage
from perun.connector.perun_openapi.model.input_get_paginated_applications import InputGetPaginatedApplications
from perun.connector.perun_openapi.model.input_get_paginated_groups import InputGetPaginatedGroups
from perun.connector.perun_openapi.model.input_get_paginated_members import InputGetPaginatedMembers
from perun.connector.perun_openapi.model.input_get_paginated_subgroups import InputGetPaginatedSubgroups
from perun.connector.perun_openapi.model.input_get_paginated_users import InputGetPaginatedUsers
from perun.connector.perun_openapi.model.input_get_resources import InputGetResources
from perun.connector.perun_openapi.model.input_get_resources1 import InputGetResources1
from perun.connector.perun_openapi.model.input_get_users import InputGetUsers
from perun.connector.perun_openapi.model.input_lock_publications import InputLockPublications
from perun.connector.perun_openapi.model.input_remove_resource_tag_from_resource import InputRemoveResourceTagFromResource
from perun.connector.perun_openapi.model.input_send_message import InputSendMessage
from perun.connector.perun_openapi.model.input_set_ban import InputSetBan
from perun.connector.perun_openapi.model.input_set_ban_for_user_on_facility import InputSetBanForUserOnFacility
from perun.connector.perun_openapi.model.input_set_facility_attribute import InputSetFacilityAttribute
from perun.connector.perun_openapi.model.input_set_facility_attributes import InputSetFacilityAttributes
from perun.connector.perun_openapi.model.input_set_facility_resource_group_user_member_attributes import InputSetFacilityResourceGroupUserMemberAttributes
from perun.connector.perun_openapi.model.input_set_facility_resource_user_member_attributes import InputSetFacilityResourceUserMemberAttributes
from perun.connector.perun_openapi.model.input_set_facility_user_attributes import InputSetFacilityUserAttributes
from perun.connector.perun_openapi.model.input_set_group_attribute import InputSetGroupAttribute
from perun.connector.perun_openapi.model.input_set_group_attributes import InputSetGroupAttributes
from perun.connector.perun_openapi.model.input_set_group_resource_attribute import InputSetGroupResourceAttribute
from perun.connector.perun_openapi.model.input_set_group_resource_attributes import InputSetGroupResourceAttributes
from perun.connector.perun_openapi.model.input_set_host_attribute import InputSetHostAttribute
from perun.connector.perun_openapi.model.input_set_host_attributes import InputSetHostAttributes
from perun.connector.perun_openapi.model.input_set_member_attribute import InputSetMemberAttribute
from perun.connector.perun_openapi.model.input_set_member_attributes import InputSetMemberAttributes
from perun.connector.perun_openapi.model.input_set_member_group_attribute import InputSetMemberGroupAttribute
from perun.connector.perun_openapi.model.input_set_member_group_attributes import InputSetMemberGroupAttributes
from perun.connector.perun_openapi.model.input_set_member_group_with_user_attributes import InputSetMemberGroupWithUserAttributes
from perun.connector.perun_openapi.model.input_set_member_resource_and_user_attributes import InputSetMemberResourceAndUserAttributes
from perun.connector.perun_openapi.model.input_set_member_resource_attribute import InputSetMemberResourceAttribute
from perun.connector.perun_openapi.model.input_set_member_resource_attributes import InputSetMemberResourceAttributes
from perun.connector.perun_openapi.model.input_set_member_with_user_attributes import InputSetMemberWithUserAttributes
from perun.connector.perun_openapi.model.input_set_resource_attribute import InputSetResourceAttribute
from perun.connector.perun_openapi.model.input_set_resource_attributes import InputSetResourceAttributes
from perun.connector.perun_openapi.model.input_set_resource_group_attributes import InputSetResourceGroupAttributes
from perun.connector.perun_openapi.model.input_set_resource_group_with_group_attributes import InputSetResourceGroupWithGroupAttributes
from perun.connector.perun_openapi.model.input_set_sending_enabled import InputSetSendingEnabled
from perun.connector.perun_openapi.model.input_set_sponsored_member import InputSetSponsoredMember
from perun.connector.perun_openapi.model.input_set_user_attribute import InputSetUserAttribute
from perun.connector.perun_openapi.model.input_set_user_attributes import InputSetUserAttributes
from perun.connector.perun_openapi.model.input_set_user_ext_source_attribute import InputSetUserExtSourceAttribute
from perun.connector.perun_openapi.model.input_set_user_ext_source_attributes import InputSetUserExtSourceAttributes
from perun.connector.perun_openapi.model.input_set_user_facility_attribute import InputSetUserFacilityAttribute
from perun.connector.perun_openapi.model.input_set_user_facility_attributes import InputSetUserFacilityAttributes
from perun.connector.perun_openapi.model.input_set_vo_attribute import InputSetVoAttribute
from perun.connector.perun_openapi.model.input_set_vo_attributes import InputSetVoAttributes
from perun.connector.perun_openapi.model.input_set_vo_ban import InputSetVoBan
from perun.connector.perun_openapi.model.input_specific_member import InputSpecificMember
from perun.connector.perun_openapi.model.input_submit_application import InputSubmitApplication
from perun.connector.perun_openapi.model.input_update_application_mail import InputUpdateApplicationMail
from perun.connector.perun_openapi.model.input_update_ban import InputUpdateBan
from perun.connector.perun_openapi.model.input_update_ban_for_facility import InputUpdateBanForFacility
from perun.connector.perun_openapi.model.input_update_category import InputUpdateCategory
from perun.connector.perun_openapi.model.input_update_consent_hub import InputUpdateConsentHub
from perun.connector.perun_openapi.model.input_update_facility import InputUpdateFacility
from perun.connector.perun_openapi.model.input_update_form import InputUpdateForm
from perun.connector.perun_openapi.model.input_update_form_items_for_group import InputUpdateFormItemsForGroup
from perun.connector.perun_openapi.model.input_update_form_items_for_vo import InputUpdateFormItemsForVo
from perun.connector.perun_openapi.model.input_update_group import InputUpdateGroup
from perun.connector.perun_openapi.model.input_update_publication import InputUpdatePublication
from perun.connector.perun_openapi.model.input_update_publication_system import InputUpdatePublicationSystem
from perun.connector.perun_openapi.model.input_update_resource import InputUpdateResource
from perun.connector.perun_openapi.model.input_update_resource_tag import InputUpdateResourceTag
from perun.connector.perun_openapi.model.input_update_service import InputUpdateService
from perun.connector.perun_openapi.model.input_update_services_package import InputUpdateServicesPackage
from perun.connector.perun_openapi.model.input_update_user import InputUpdateUser
from perun.connector.perun_openapi.model.input_update_vo import InputUpdateVo
from perun.connector.perun_openapi.model.item_texts import ItemTexts
from perun.connector.perun_openapi.model.mail_text import MailText
from perun.connector.perun_openapi.model.mail_type import MailType
from perun.connector.perun_openapi.model.member import Member
from perun.connector.perun_openapi.model.member_all_of import MemberAllOf
from perun.connector.perun_openapi.model.member_candidate import MemberCandidate
from perun.connector.perun_openapi.model.member_group_status import MemberGroupStatus
from perun.connector.perun_openapi.model.member_with_sponsors import MemberWithSponsors
from perun.connector.perun_openapi.model.members_order_column import MembersOrderColumn
from perun.connector.perun_openapi.model.members_page_query import MembersPageQuery
from perun.connector.perun_openapi.model.namespace_rules import NamespaceRules
from perun.connector.perun_openapi.model.new_apps import NewApps
from perun.connector.perun_openapi.model.owner import Owner
from perun.connector.perun_openapi.model.owner_all_of import OwnerAllOf
from perun.connector.perun_openapi.model.paginated_audit_messages import PaginatedAuditMessages
from perun.connector.perun_openapi.model.paginated_rich_applications import PaginatedRichApplications
from perun.connector.perun_openapi.model.paginated_rich_groups import PaginatedRichGroups
from perun.connector.perun_openapi.model.paginated_rich_members import PaginatedRichMembers
from perun.connector.perun_openapi.model.paginated_rich_users import PaginatedRichUsers
from perun.connector.perun_openapi.model.perun_apps_config import PerunAppsConfig
from perun.connector.perun_openapi.model.perun_bean import PerunBean
from perun.connector.perun_openapi.model.perun_exception import PerunException
from perun.connector.perun_openapi.model.perun_policy import PerunPolicy
from perun.connector.perun_openapi.model.perun_principal import PerunPrincipal
from perun.connector.perun_openapi.model.publication import Publication
from perun.connector.perun_openapi.model.publication_all_of import PublicationAllOf
from perun.connector.perun_openapi.model.publication_for_gui import PublicationForGUI
from perun.connector.perun_openapi.model.publication_for_gui_all_of import PublicationForGUIAllOf
from perun.connector.perun_openapi.model.publication_system import PublicationSystem
from perun.connector.perun_openapi.model.publication_system_all_of import PublicationSystemAllOf
from perun.connector.perun_openapi.model.rt_message import RTMessage
from perun.connector.perun_openapi.model.resource import Resource
from perun.connector.perun_openapi.model.resource_all_of import ResourceAllOf
from perun.connector.perun_openapi.model.resource_state import ResourceState
from perun.connector.perun_openapi.model.resource_tag import ResourceTag
from perun.connector.perun_openapi.model.resource_tag_all_of import ResourceTagAllOf
from perun.connector.perun_openapi.model.rich_application import RichApplication
from perun.connector.perun_openapi.model.rich_application_all_of import RichApplicationAllOf
from perun.connector.perun_openapi.model.rich_destination import RichDestination
from perun.connector.perun_openapi.model.rich_destination_all_of import RichDestinationAllOf
from perun.connector.perun_openapi.model.rich_facility import RichFacility
from perun.connector.perun_openapi.model.rich_facility_all_of import RichFacilityAllOf
from perun.connector.perun_openapi.model.rich_group import RichGroup
from perun.connector.perun_openapi.model.rich_group_all_of import RichGroupAllOf
from perun.connector.perun_openapi.model.rich_member import RichMember
from perun.connector.perun_openapi.model.rich_member_all_of import RichMemberAllOf
from perun.connector.perun_openapi.model.rich_resource import RichResource
from perun.connector.perun_openapi.model.rich_resource_all_of import RichResourceAllOf
from perun.connector.perun_openapi.model.rich_user import RichUser
from perun.connector.perun_openapi.model.rich_user_all_of import RichUserAllOf
from perun.connector.perun_openapi.model.rich_user_ext_source import RichUserExtSource
from perun.connector.perun_openapi.model.rich_user_ext_source_all_of import RichUserExtSourceAllOf
from perun.connector.perun_openapi.model.role_management_rules import RoleManagementRules
from perun.connector.perun_openapi.model.role_object import RoleObject
from perun.connector.perun_openapi.model.security_team import SecurityTeam
from perun.connector.perun_openapi.model.service import Service
from perun.connector.perun_openapi.model.service_all_of import ServiceAllOf
from perun.connector.perun_openapi.model.service_attributes import ServiceAttributes
from perun.connector.perun_openapi.model.service_for_gui import ServiceForGUI
from perun.connector.perun_openapi.model.service_for_gui_all_of import ServiceForGUIAllOf
from perun.connector.perun_openapi.model.service_state import ServiceState
from perun.connector.perun_openapi.model.services_package import ServicesPackage
from perun.connector.perun_openapi.model.services_package_all_of import ServicesPackageAllOf
from perun.connector.perun_openapi.model.set_role_with_group_complementary_object import SetRoleWithGroupComplementaryObject
from perun.connector.perun_openapi.model.set_role_with_user_complementary_object import SetRoleWithUserComplementaryObject
from perun.connector.perun_openapi.model.simple_attribute import SimpleAttribute
from perun.connector.perun_openapi.model.sorting_order import SortingOrder
from perun.connector.perun_openapi.model.sponsor import Sponsor
from perun.connector.perun_openapi.model.sponsored_user_data import SponsoredUserData
from perun.connector.perun_openapi.model.task import Task
from perun.connector.perun_openapi.model.task_and_destination_id_object import TaskAndDestinationIdObject
from perun.connector.perun_openapi.model.task_and_destination_name_object import TaskAndDestinationNameObject
from perun.connector.perun_openapi.model.task_id_object import TaskIdObject
from perun.connector.perun_openapi.model.task_result import TaskResult
from perun.connector.perun_openapi.model.task_result_all_of import TaskResultAllOf
from perun.connector.perun_openapi.model.task_result_id_object import TaskResultIdObject
from perun.connector.perun_openapi.model.task_result_status import TaskResultStatus
from perun.connector.perun_openapi.model.task_status import TaskStatus
from perun.connector.perun_openapi.model.thanks import Thanks
from perun.connector.perun_openapi.model.thanks_all_of import ThanksAllOf
from perun.connector.perun_openapi.model.thanks_for_gui import ThanksForGUI
from perun.connector.perun_openapi.model.thanks_for_gui_all_of import ThanksForGUIAllOf
from perun.connector.perun_openapi.model.type import Type
from perun.connector.perun_openapi.model.unset_role_with_group_complementary_object import UnsetRoleWithGroupComplementaryObject
from perun.connector.perun_openapi.model.unset_role_with_user_complementary_object import UnsetRoleWithUserComplementaryObject
from perun.connector.perun_openapi.model.user import User
from perun.connector.perun_openapi.model.user_all_of import UserAllOf
from perun.connector.perun_openapi.model.user_ext_source import UserExtSource
from perun.connector.perun_openapi.model.user_ext_source_all_of import UserExtSourceAllOf
from perun.connector.perun_openapi.model.users_order_column import UsersOrderColumn
from perun.connector.perun_openapi.model.users_page_query import UsersPageQuery
from perun.connector.perun_openapi.model.vo import Vo
from perun.connector.perun_openapi.model.vo_admin_roles import VoAdminRoles
from perun.connector.perun_openapi.model.vo_all_of import VoAllOf
from perun.connector.perun_openapi.model.vo_member_statuses import VoMemberStatuses
